/*
 * kgraft_patch_bsc1012852
 *
 * Fix for CVE-2016-8632, bsc#1012852
 *
 *  Upstream commit:
 *  3de81b758853 ("tipc: check minimum bearer MTU")
 *
 *  SLE12(-SP1) commit:
 *  e486ebe4692c819035035eed96f8963c9e6ee965
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>

#include "local.h"

#define KGR_PATCHED_MODULE      "tipc"

static struct eth_media (*kgr_eth_media_array)[MAX_ETH_MEDIA];
static int (*kgr_recv_msg)(struct sk_buff *buf, struct net_device *dev, struct packet_type *pt, struct net_device *orig_dev);
static void (*kgr_setup_media)(struct work_struct *work);
static void (*kgr_eth_media_addr_set)(const struct tipc_bearer *tb_ptr, struct tipc_media_addr *a, char *mac);
static int (*kgr_tipc_reset_bearer)(struct tipc_bearer *b_ptr);
static int (*kgr_tipc_disable_bearer)(const char *name);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "eth_media_array", (void *)&kgr_eth_media_array },
	{ "tipc:recv_msg", (void *)&kgr_recv_msg },
	{ "tipc:setup_media", (void *)&kgr_setup_media },
	{ "eth_media_addr_set", (void *)&kgr_eth_media_addr_set },
	{ "tipc_reset_bearer", (void *)&kgr_tipc_reset_bearer },
	{ "tipc_disable_bearer", (void *)&kgr_tipc_disable_bearer },
};

/* patched function */
int kgr_enable_media(struct tipc_bearer *tb_ptr)
{
	struct net_device *dev;
	struct eth_media *eb_ptr = &(*kgr_eth_media_array)[0];
	struct eth_media *stop = &(*kgr_eth_media_array)[MAX_ETH_MEDIA];
	char *driver_name = strchr((const char *)tb_ptr->name, ':') + 1;
	int pending_dev = 0;

	/* Find unused Ethernet bearer structure */
	while (eb_ptr->dev) {
		if (!eb_ptr->bearer)
			pending_dev++;
		if (++eb_ptr == stop)
			return pending_dev ? -EAGAIN : -EDQUOT;
	}

	/* Find device with specified name */
	dev = dev_get_by_name(&init_net, driver_name);
	if (!dev)
		return -ENODEV;
	/*
	 * Fix CVE-2016-8632
	 *  +4 lines
	 */
	if (kgr_tipc_mtu_bad(dev, 0)) {
		dev_put(dev);
		return -EINVAL;
	}

	/* Create Ethernet bearer for device */
	eb_ptr->dev = dev;
	eb_ptr->tipc_packet_type.type = htons(ETH_P_TIPC);
	eb_ptr->tipc_packet_type.dev = dev;
	eb_ptr->tipc_packet_type.func = kgr_recv_msg;
	eb_ptr->tipc_packet_type.af_packet_priv = eb_ptr;
	INIT_LIST_HEAD(&(eb_ptr->tipc_packet_type.list));
	INIT_WORK(&eb_ptr->setup, kgr_setup_media);
	schedule_work(&eb_ptr->setup);

	/* Associate TIPC bearer with Ethernet bearer */
	eb_ptr->bearer = tb_ptr;
	tb_ptr->usr_handle = (void *)eb_ptr;
	memset(tb_ptr->bcast_addr.value, 0, sizeof(tb_ptr->bcast_addr.value));
	memcpy(tb_ptr->bcast_addr.value, dev->broadcast, ETH_ALEN);
	tb_ptr->bcast_addr.media_id = TIPC_MEDIA_TYPE_ETH;
	tb_ptr->bcast_addr.broadcast = 1;
	tb_ptr->mtu = dev->mtu;
	kgr_eth_media_addr_set(tb_ptr, &tb_ptr->addr, (char *)dev->dev_addr);
	return 0;
}

/* patched function */
int kgr_recv_notification(struct notifier_block *nb, unsigned long evt, void *ptr)
{
	struct net_device *dev = netdev_notifier_info_to_dev(ptr);
	struct eth_media *eb_ptr = &(*kgr_eth_media_array)[0];
	struct eth_media *stop = &(*kgr_eth_media_array)[MAX_ETH_MEDIA];

	if (!net_eq(dev_net(dev), &init_net))
		return NOTIFY_DONE;

	while ((eb_ptr->dev != dev)) {
		if (++eb_ptr == stop)
			return NOTIFY_DONE;	/* couldn't find device */
	}
	if (!eb_ptr->bearer)
		return NOTIFY_DONE;		/* bearer had been disabled */

	/*
	 * Fix CVE-2016-8632
	 *  -eb_ptr->bearer->mtu = dev->mtu;
	 *  -
	 */
	switch (evt) {
	case NETDEV_CHANGE:
		if (netif_carrier_ok(dev))
			break;
	case NETDEV_DOWN:
	/*
	 * Fix CVE-2016-8632
	 *  -case NETDEV_CHANGEMTU:
	 */
	case NETDEV_CHANGEADDR:
		kgr_tipc_reset_bearer(eb_ptr->bearer);
		break;
	/*
	 * Fix CVE-2016-8632
	 *  +8 lines
	 */
	case NETDEV_CHANGEMTU:
		if (kgr_tipc_mtu_bad(dev, 0)) {
			kgr_tipc_disable_bearer(eb_ptr->bearer->name);
			break;
		}
		eb_ptr->bearer->mtu = dev->mtu;
		kgr_tipc_reset_bearer(eb_ptr->bearer);
		break;
	case NETDEV_UNREGISTER:
	case NETDEV_CHANGENAME:
		kgr_tipc_disable_bearer(eb_ptr->bearer->name);
		break;
	}
	return NOTIFY_OK;
}

static int kgr_patch_bsc1012852_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int kgr_patch_bsc1012852_module_notify(struct notifier_block *nb,
	unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1012852_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1012852_module_nb = {
	.notifier_call = kgr_patch_bsc1012852_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1012852_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1012852_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1012852_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1012852_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1012852_module_nb);
}
